/*
 * Copyright (c) 1997 Loughborough University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the LUTCHI Research
 *      Centre at Loughborough University.
 * 4. Neither the name of the University nor of the Centre may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


/*
 *  File: main.c
 *
 *  Description: Whiteboard initialisation and main loop.
 *
 *  J.C.Highfield, 3/97.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <tcl.h>
#include <tk.h>

#include "wb.h"
#include "db.h"
#include "proto.h"

#ifndef EXT_SCRIPT
extern char code[];
#endif

Tcl_Interp *tcl;
Tk_Window MainWin;

int help       = 0;
int maxhops    = 5;
int orient     = 1;
int readonly   = 0;
int pssize     = 32768;
int datarate   = 32768 * 8;
char *username = "";
char *confname = "Whiteboard";
char *geometry = "";

int   confPort       = 2323;
char  confGroup[32]  = "230.230.230.230";


Tk_ArgvInfo argTable[] = {
    {"-help", TK_ARGV_CONSTANT, (char *) 1, (char *) &help,
	"Show command usage information"},
    {"-N", TK_ARGV_STRING, NULL, (char *) &username,
        "Username"},
    {"-C", TK_ARGV_STRING, NULL, (char *) &confname,
        "Conference name"},
    {"-maxhops", TK_ARGV_INT, NULL, (char *) &maxhops,
	"Maximum TTL for advertising sessions"},
    {"-t", TK_ARGV_INT, NULL, (char *) &maxhops,
	"Maximum TTL for advertising sessions"},
    {"-p", TK_ARGV_CONSTANT, (char *) 0, (char *) &orient,
	"Portrait mode"},
    {"-l", TK_ARGV_CONSTANT, (char *) 1, (char *) &orient,
	"Landscape mode"},
    {"+l", TK_ARGV_CONSTANT, (char *) 3, (char *) &orient,
	"Landscape mode"},
    {"-r", TK_ARGV_CONSTANT, (char *) 1, (char *) &readonly,
	"Read only"},
    {"+r", TK_ARGV_CONSTANT, (char *) 0, (char *) &readonly,
	"Write enabled"},
    {"-geometry", TK_ARGV_STRING, NULL, (char *) &geometry,
        "Window geometry"},
    {"-P", TK_ARGV_INT, NULL, (char *) &pssize,
	"Maximum Postscript file size allowed"},
    {"-R", TK_ARGV_INT, NULL, (char *) &datarate,
	"Maximum network data rate (bits/second)"},
    {NULL, TK_ARGV_END, NULL, NULL, NULL}
};

/* Global. */
int rsock;
int tsock;

/* Resize global. */
long resizeNeeded = 0;

#ifdef STAND_ALONE
void *TkSetPlatformInit(int (*func)(Tcl_Interp *));
extern void TkCreateXEventSource(void);

int TkPlatformInit (Tcl_Interp *interp)
{
	Tcl_SetVar(interp, "tk_library", ".", TCL_GLOBAL_ONLY);
	TkCreateXEventSource();

	return (TCL_OK);
}
#endif

void usage (void)
{
    printf ("Usage: wbd [-N username] [-C confname] [-maxhops n]\n");
    printf ("           [-p|-l|+l] [-r|+r] [-geometry wxh+x+y]  \n");
    printf ("           [-P filesize] [-R bandwidth] [-help] address/port\n");
}

void moreusage (void)
{
    printf ("\nUsage: wbd [-N username] [-C confname] [-maxhops n]\n");
    printf ("           [-p|-l|+l] [-r|+r] [-geometry wxh+x+y]  \n");
    printf ("           [-P filesize] [-R bandwidth] [-help] address/port\n");
    printf ("    -N username  Set the announced username.\n");
    printf ("    -C confname  Set the conference window name.\n");
    printf ("    -maxhops n   Set the TTL to n.\n");
    printf ("    -t n         Set the TTL to n, (as -maxhops).\n");
    printf ("    -p,-l,+l     Portrait/Landscape mode,\n");
    printf ("    -r,+r        Disable/Enable writing to the whiteboard.\n");
    printf ("    -geometry g  Set the window's geometry to g (wxh+x+y).\n");
    printf ("    -P size      Set the Postscript file size limit to size.\n");
    printf ("    -R bw        Set the bandwidth limit to bw (bits/s).\n");
    printf ("    -help        Display this message.\n\n");
}

void tkStringSet (const char *name, const char *value)
{
    char arg[1024];

    sprintf (arg, "set W(%s) \"%s\"", name, value);
    Tcl_Eval (tcl, arg);
}

void tkIntegerSet (const char *name, const int value)
{
    char arg[1024];

    sprintf (arg, "set W(%s) %d", name, value);
    Tcl_Eval (tcl, arg);
}

int main (int argc, char **argv)
{
    int  length;
    int  result;
    int  i, port;
    char *cmd;
    char *p;
    char arg[1024] = "";
    char host[20] = "";
    uint8  buffer[2048];
    struct sockaddr_in addr;

    /* Parse the command line. */
    tcl = Tcl_CreateInterp();
    cmd = Tcl_Concat(argc, argv);

    result = Tk_ParseArgv(tcl, NULL, &argc, argv, argTable,
	TK_ARGV_NO_DEFAULTS);
    if ((result != TCL_OK) || (argc > 2))
    {
        usage();
        exit(-1);
    }
    if (help)
    {
        moreusage();
        exit (0);
    }

    /* Parse remaining argument. */
    if (argc > 1)
    {
        if ((p = strchr (argv[1], '/')) != NULL)
        {
            *p++ = '\0';
            port = -1;
            sscanf (p, "%d", &confPort);
        }

        strcpy (confGroup, argv[1]);
    }

    /* Check some settings for validity. */
    if (pssize > (1024*1024))
        pssize = (1024*1024);
    if (datarate > (512*1024))
        datarate = (512*1024);
    if (datarate < 8192)
        datarate = 8192;

    /* Initialise Tcl and Tk. */
#ifndef STAND_ALONE
    if (Tcl_Init (tcl) != TCL_OK)
    {
        printf ("Failed to initialise tcl: %s\n", tcl->result);
        exit (-1);
    }
#endif
    if (Tk_Init (tcl) != TCL_OK)
    {
        printf ("Failed to initialise tk: %s\n", tcl->result);
        exit (-1);
    }

    Tcl_StaticPackage(tcl, "Tk", Tk_Init, (Tcl_PackageInitProc *) NULL);

    MainWin = Tk_MainWindow(tcl);

    /* Create commands. */
    Tcl_CreateCommand(tcl, "Draw", Draw, (ClientData) NULL,
        (void (*)()) NULL);
    Tcl_CreateCommand(tcl, "UnDelete", UnDelete, (ClientData) NULL,
        (void (*)()) NULL);
    Tcl_CreateCommand(tcl, "SendPS", SendPS, (ClientData) NULL,
        (void (*)()) NULL);
    Tcl_CreateCommand(tcl, "PageNew", PageNew, (ClientData) NULL,
        (void (*)()) NULL);
    Tcl_CreateCommand(tcl, "PageSet", PageSetTcl, (ClientData) NULL,
        (void (*)()) NULL);
    Tcl_CreateCommand(tcl, "PageCanvasSize", PageCanvasSize, (ClientData) NULL,
        (void (*)()) NULL);
    Tcl_CreateCommand(tcl, "WbDump", WbDump, (ClientData) NULL,
        (void (*)()) NULL);
    Tcl_CreateCommand(tcl, "WbReload", WbReload, (ClientData) NULL,
        (void (*)()) NULL);
    Tcl_CreateCommand(tcl, "Sender", Sender, (ClientData) NULL,
        (void (*)()) NULL);

    /* Pass command line variables to tcl. */
    if (strlen(username) > 0)
        ascii_source (username);
    tkStringSet ("username",  ascii_source(NULL));
    tkStringSet ("confname",  confname);
    tkStringSet ("confGroup", confGroup);
    tkIntegerSet ("confPort", confPort);
    tkIntegerSet ("confTTL",  maxhops);
    tkStringSet  ("geometry", geometry);
    tkIntegerSet ("orientation", orient);
    tkIntegerSet ("readonly", readonly);
    tkIntegerSet ("pssize", pssize);

    /* Run the initialisation script. */
#ifdef EXT_SCRIPT
    if (Tcl_VarEval(tcl, "source ", "wbd.tcl", (char *) NULL) != TCL_OK)
    {
        printf ("Failed to run initialisation script!\n");
        printf ("  Result: %s\n", tcl->result);
        exit (-1);
    }
#else
    if (Tcl_Eval(tcl, (char *)code) != TCL_OK)
    {
        printf ("Failed to run compiled initialisation script!\n");
        printf ("  Result: %s\n", tcl->result);
        exit (-1);
    }
#endif

    /* Create network name with wildcards. */
    memset (&addr, 0, sizeof (struct sockaddr_in));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = inet_addr (confGroup);
    addr.sin_port = htons(confPort);

    /* Open a multicast socket for receiving. */
    rsock = NetInitReceive(&addr);
    if (rsock < 0)
    {
        perror ("no receive socket");
        exit (-1);
    }

    /* Open a multicast socket for transmitting. */
    tsock = NetInitTransmit(&addr);
    if (tsock < 0)
    {
        perror ("no transmit socket");
        exit (-1);
    }

    /* Install the window resize handler. */
    Tk_CreateEventHandler(MainWin, StructureNotifyMask,
	    ResizeProc, (ClientData) NULL);

    /* Note initial sender name. */
    SenderAdd (this_ip(), time_stamp(), time_stamp(), 
                        PK_SESSION, ascii_source(NULL));

    /* Run round in circles; Display windows! */
    while (1)
    {
        if (resizeNeeded == 1)
        {
            resizeNeeded = 0;
            DoResize();
        }

        if (Tcl_DoOneEvent(TCL_DONT_WAIT) == 0)
        {
            /* If idle, take a nap. */
#ifdef SGI_NAP
            sginap(2);        /* 0.02s */
#else
            usleep(20000);    /* 0.02s */
#endif
        }

        while (Tcl_DoOneEvent(TCL_DONT_WAIT) != 0)
            ;

        /* Update timer and send queues. */
        CheckQueues();

        Tcl_ResetResult (tcl);
        length = NetFetch(rsock, &addr, buffer, 2048);
        if (length > 0)
            NetParse(buffer, length);
    }

    return 0;
}


/* When the user resizes the window take a note */
/* and do it later - avoid reentrancy problems. */
void ResizeProc (ClientData clientData, XEvent *eventPtr)
{
    if (eventPtr->type != ConfigureNotify)
        return;

    resizeNeeded = 1;
}


void DoResize (void)
{
    char arg[1024];
    extern Tcl_Interp *tcl;
    Tk_Window edge, info;
    int h, w;

    edge   = Tk_NameToWindow (tcl, ".edge",  MainWin);
    info   = Tk_NameToWindow (tcl, ".info",  MainWin);

    h = Tk_Height (MainWin) - Tk_ReqHeight (info) - 4;
    w = Tk_Width (MainWin)  - Tk_ReqWidth (edge)  - 4;

/*  printf ("%d %d::%d %d\n", Tk_Height (MainWin),
         Tk_Width (MainWin), Tk_ReqHeight (info), 
                              Tk_ReqWidth (edge));*/

    sprintf (arg, "CanvasResize %d %d", w, h);
    if (Tcl_Eval (tcl, arg) != TCL_OK)
        printf ("%s RESULT %s\n", arg, tcl->result);

    sprintf (arg, "update");
    if (Tcl_Eval (tcl, arg) != TCL_OK)
        printf ("%s RESULT %s\n", arg, tcl->result);
}

